unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, TeeGDIPlus, TeEngine, ExtCtrls, TeeProcs, Chart, TeeTools, Series,
  StdCtrls, Menus, Math, System.Generics.Collections;

{$define CreateBar}

type
  TColorBandTools = TObjectList<TColorBandTool>;
  TAnnotationTools = TObjectList<TAnnotationTool>;

  TForm1 = class(TForm)
    Chart1: TChart;
    Panel1: TPanel;
    chkUseCalloutPos: TCheckBox;
    Panel2: TPanel;
    lbxEvents: TListBox;
    Panel3: TPanel;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    lblMouseDown: TLabel;
    lblMouseMove: TLabel;
    lblMouseUp: TLabel;
    btnClearBands: TButton;
    Label4: TLabel;
    lblStartIdx: TLabel;
    Label6: TLabel;
    lblEndIdx: TLabel;
    Timer1: TTimer;
    Label5: TLabel;
    lblChartMousePos: TLabel;
    Series1: TBarSeries;
    Label7: TLabel;
    lblEndValue: TLabel;
    ChartTool1: TAnnotationTool;
    procedure btnClearBandsClick(Sender: TObject);
    procedure Chart1BeforeDrawSeries(Sender: TObject);
    procedure Chart1MouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure Chart1MouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure Chart1MouseUp(Sender: TObject; Button: TMouseButton; Shift:
        TShiftState; X, Y: Integer);
    procedure Timer1Timer(Sender: TObject);
    procedure cxButton1Click(Sender: TObject);
  private
    FBands: TColorBandTools;
    FAnnotations: TAnnotationTools;
    FActiveColorBand: TColorBandTool;
    FActiveAnnotation: TAnnotationTool;
    function CreateColorBand: TColorBandTool;
    function CreateAnnotation: TAnnotationTool;
    function GetAnnCaption(const aIndex: integer): string;
    procedure AddTimeBand;
    //function GetXPosFromValue(const aXValue: double): integer;
    //function GetLastBandXValue: integer;
    function GetValueIndex(const aXPos: integer): integer;
    procedure CheckListBox;
    procedure ClearMouseTimes;
    function CalcOffset(const index: Integer; aXPos: Integer): Double;
    procedure UpdateAnnotations(aChart: TChart);
    procedure UpdateAnnPos(const aAnnTool: TAnnotationTool; const aColorBand: TColorBandTool);
    { Private declarations }
  public
    { Public declarations }
    constructor Create(aOwner: TComponent); override;
    destructor Destroy; override;
  end;

var
  Form1: TForm1;
implementation

{$if defined(CodeSite)}
uses
  CodeSiteLogging;
{$endif}

{$R *.dfm}

type
  TAnnotationAccess = class(TAnnotationTool);

const
  COLOR_BAND_DIFF = 0.0001;

function TrimTrailingZeros(const aStr: String): string;
var
  lDecPos: Integer;
begin
  result := aStr;
  lDecPos := Pos('.', result);
  if lDecPos = 0 then
    exit;

  while (Length(result) > lDecPos) and (result[Length(result)] = '0') do
    Delete(result, Length(result), 1);

  if (Length(result) = lDecPos) then
    Delete(result, Length(result), 1);
end;

function TrimmedValue(const aValue: double; const aMaxZeros: integer = 4): string;
begin
  result := TrimTrailingZeros(Format('%.' + IntToStr(aMaxZeros)+ 'f', [aValue]));
end;

function TForm1.GetValueIndex(const aXPos: integer): integer;
var
  I,
  lDistance: integer;

  procedure UpdateMinDistance(const aValueIndex: Integer);
  var
    lXPos1, lXPos2: integer;
    lCalcedDist: integer;
  begin
    lXPos1:= Series1.CalcXPos(aValueIndex);
    lXPos2:= lXPos1 + Series1.BarWidth;
    lCalcedDist := Min(abs(aXPos - lXPos1), abs(aXPos - lXPos2));

    if (lDistance < 0) or (lDistance > lCalcedDist) then
    begin
      lDistance:= lCalcedDist;
      result := aValueIndex;
    end;
  end;

begin
  Assert(Series1 is TBarSeries);

  result := - 1;

  if (Series1.Count > 0) then
  begin
    lDistance := -1;
    UpdateMinDistance(0);
    for I := 1 to Series1.Count-1 do
      UpdateMinDistance(I);
  end;
end;

function TForm1.GetAnnCaption(const aIndex: integer): string;
begin
   result := ' Alloc. #' + inttostr(aIndex + 1) + ' ';
end;

function TForm1.CreateAnnotation: TAnnotationTool;
begin
  result := TAnnotationTool.Create(self);
  result.ParentChart := Chart1;
  result.Shape.CustomPosition := True;
  result.Shape.Font.Height := -13;
  result.Shape.Left := 440;
  result.Shape.Margins.Left := 20;
  result.Shape.Margins.Top := 10;
  result.Shape.Margins.Right := 20;
  result.Shape.Margins.Bottom := 10;
  result.Shape.ShapeStyle := fosRoundRectangle;
  result.Shape.Text := GetAnnCaption(FAnnotations.Count);
  result.Shape.Top := 152;
end;

function TForm1.CreateColorBand: TColorBandTool;
begin
  result := TColorBandTool.Create(self);
  result.ParentChart := Chart1;
  result.Axis := Chart1.BottomAxis;
  result.Pen.Visible := False;
  result.StartLinePen.Color := clBlack;
  result.EndLinePen.Color := clBlack;
  result.DrawBehind := True;
  result.AllowDrag := False;
  result.ResizeStart := false;
  result.ResizeEnd := false;
  result.Transparency := 50;
end;

procedure TForm1.Chart1MouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  lIdx: Integer;
  lOffset: double;
  lStartValue: Double;
begin
  lblMouseDown.Caption := TimeToStr(Now);
  Caption := 'Down: ' + TimeToStr(Now);

  Assert(not (Assigned(FActiveColorBand) xor assigned(FActiveAnnotation)));

  lIdx := GetValueIndex(X);
  if (not assigned(FActiveColorBand)) and (lIdx > -1) then
  begin
    lOffset := CalcOffset(lIdx, X);

    lStartValue := Series1.XValue[lIdx] + lOffset;
    if (FBands.Count > 0) and (FBands.Last.EndValue > lStartValue) then
      exit;

    FActiveColorBand := CreateColorBand;
    FActiveAnnotation := CreateAnnotation;
    FActiveAnnotation.Tag := integer(FActiveColorBand);
    FActiveColorBand.Tag := integer(FActiveAnnotation);

    if (FBands.Count = 0) then
      FActiveColorBand.StartValue := 0
    else
      FActiveColorBand.StartValue := FBands.Last.EndValue;

    {$if defined(CodeSite)}
    CodeSite.Write('EndValue', 'MouseDown');
    {$endif}
    FActiveColorBand.EndValue := Series1.XValue[lIdx] + lOffset;
    FActiveColorBand.Visible := True;

    UpdateAnnPos(FActiveAnnotation, FActiveColorBand);

    FBands.Add(FActiveColorBand);
    FAnnotations.Add(FActiveAnnotation);
  end;

  Assert(not (Assigned(FActiveColorBand) xor assigned(FActiveAnnotation)));

  if assigned(FActiveColorBand) then
  begin
    lbxEvents.Items.Insert(0, 'MouseDown');
    CheckListBox;
  end;
end;

constructor TForm1.Create(aOwner: TComponent);
var
  I: Integer;
  lValue: double;
begin
  inherited;


  FBands := TColorBandTools.Create;
  FAnnotations := TAnnotationTools.Create;

  FActiveColorBand := nil;
  FActiveAnnotation := nil;

  Chart1.View3D:=false;

  Chart1.FreeAllSeries;
  Series1 := TBarSeries.Create(self);
  Series1.Marks.Visible := false;
  Chart1.AddSeries(Series1);

  for I := Chart1.Tools.Count-1 downto 0 do
    if (Chart1.Tools[I] is TColorBandTool) or (Chart1.Tools[I] is TAnnotationTool) then
      Chart1.Tools[I].Free;



  lValue := 1000;
  for I := 1 to 25 do
  begin
    Series1.AddXY(I, lValue);
    lValue := lValue * 1.03;
  end;


  ClearMouseTimes;
end;

procedure TForm1.cxButton1Click(Sender: TObject);
begin
  AddTimeBand;
end;

destructor TForm1.Destroy;
begin
  FBands.Free;
  FAnnotations.Free;
  inherited;
end;

procedure TForm1.AddTimeBand;
{var
  lIdx: Integer;
  lPt: TPoint;}
begin
{  lIdx := GetValueIndex(GetLastBandXValue);
  if lIdx > -1 then
  begin
    lPt.X := GetXPosFromValue(GetLastBandXValue);
    lPt.Y := Chart1.LeftAxis.CalcYPosValue(Chart1.LeftAxis.Maximum * 0.5);
    lPt := Chart1.ClientToScreen(lPt);

    SetCursorPos(lPt.X + 1, lPt.Y);
    cbHighlight.Visible := True;
  end
  else
    cbHighlight.Visible := false;

  if cbHighlight.Visible then
  begin
    lbxEvents.Items.Insert(0, 'MouseDown');
    CheckListBox;
  end;
}
end;

procedure TForm1.btnClearBandsClick(Sender: TObject);
begin
  FBands.Clear;
  FAnnotations.Clear;
end;

procedure TForm1.ClearMouseTimes;
begin
  lblMouseDown.Caption := '-';
  lblMouseMove.Caption := '-';
  lblMouseUp.Caption := '-';
end;
procedure TForm1.Chart1MouseMove(Sender: TObject; Shift: TShiftState; X, Y:
    Integer);
var
  lIdx: Integer;
  lText: string;
begin
  lblChartMousePos.Caption := Format('%d, %d', [X, Y]);

  Assert(not (Assigned(FActiveColorBand) xor assigned(FActiveAnnotation)));

  if assigned(FActiveColorBand) then
  begin
    lblMouseMove.Caption := TimeToStr(Now);
    lText := Format('MouseMove: X: %d; Y: %d', [X, Y]);
    if (lbxEvents.Items.Count > 0) and
      (lbxEvents.Items.Objects[0] = TObject(1)) then
      lbxEvents.Items[0] := lText
    else
      lbxEvents.Items.InsertObject(0, 'MouseMove', TObject(1));
    CheckListBox;

    lIdx := GetValueIndex(X);
    if lIdx > -1 then
    begin
      {$if defined(CodeSite)}
      CodeSite.Write('Series1.XValue[lIdx]', Series1.XValue[lIdx]);
      CodeSite.Write('CalcOffset(lIdx, X)', CalcOffset(lIdx, X));
      {$endif}

      FActiveColorBand.EndValue := Series1.XValue[lIdx] + CalcOffset(lIdx, X);
      {$if defined(CodeSite)}
      CodeSite.Write('EndValue', 'MouseMove');
      CodeSite.Write('FActiveColorBand.EndValue', FActiveColorBand.EndValue);
      {$endif}
      UpdateAnnPos(FActiveAnnotation, FActiveColorBand);
    end;
  end;

Assert(not (Assigned(FActiveColorBand) xor assigned(FActiveAnnotation)));
end;

procedure TForm1.Chart1MouseUp(Sender: TObject; Button: TMouseButton; Shift:
    TShiftState; X, Y: Integer);
begin
  Assert(not (Assigned(FActiveColorBand) xor assigned(FActiveAnnotation)));

  if assigned(FActiveColorBand) then
  begin
    if (FActiveColorBand.StartValue > (FActiveColorBand.EndValue - COLOR_BAND_DIFF)) then
    begin
      FBands.Delete(FBands.IndexOf(FActiveColorBand));
      FAnnotations.Delete(FAnnotations.IndexOf(FActiveAnnotation));
    end;

    lblMouseUp.Caption := TimeToStr(Now);
    lbxEvents.Items.Insert(0, 'MouseUp');

    FActiveColorBand := nil;
    FActiveAnnotation := nil;
  end;
  Assert(not (Assigned(FActiveColorBand) xor assigned(FActiveAnnotation)));
end;

procedure TForm1.CheckListBox;
begin
  while lbxEvents.Items.Count > 20 do
    lbxEvents.Items.Delete(lbxEvents.Items.Count-1);
end;

procedure TForm1.Timer1Timer(Sender: TObject);
begin
  if assigned(FActiveColorBand) then
  begin
    lblStartIdx.Caption := IntToStr(Trunc(FActiveColorBand.StartValue));
    lblEndIdx.Caption := IntToStr(Trunc(FActiveColorBand.EndValue));
    lblEndValue.Caption := TrimmedValue(FActiveColorBand.EndValue);
  end
  else
  begin
    lblStartIdx.Caption := '-';
    lblEndIdx.Caption := '-';
    lblEndValue.Caption := '-';
  end;
end;

procedure TForm1.UpdateAnnotations(aChart: TChart);
  function GetColorBandTool(const aStartValue: double): TColorBandTool;
  var
    I: Integer;
  begin
    for I := 0 to Chart1.Tools.Count-1 do
      if (Chart1.Tools[i] is TColorBandTool) then
      begin
        result := TColorBandTool(Chart1.Tools[i]);
        if SameValue(result.StartValue, aStartValue, COLOR_BAND_DIFF) then
          exit(result);
      end;
    result := nil;
  end;

  function GetAnnotationTool(const aCaption: string): TAnnotationTool;
  var
    I: Integer;
  begin
    for I := 0 to Chart1.Tools.Count-1 do
      if (Chart1.Tools[i] is TAnnotationTool) then
      begin
        result := TAnnotationTool(Chart1.Tools[i]);
        if SameText(result.Text, aCaption) then
          exit(result);
      end;
    result := nil;
  end;

  function GetAnnoCaptions: string;
  var
    I: Integer;
  begin
    result := '';
    for I := 0 to Chart1.Tools.Count-1 do
      if (Chart1.Tools[i] is TAnnotationTool) then
        result := result + TAnnotationTool(Chart1.Tools[i]).Text + '; ';
  end;

var
  lIdx: Integer;
  lColorBand: TColorBandTool;
  lAnnTool: TAnnotationTool;
  lStartValue: Double;
begin
{$ifdef CodeSite}
if CSLogging then CodeSite.TraceMethod  (self, 'TframeStretchIRAChart.UpdateAnnotationsitions');
{$endif}
  // find color bands, and then find annotation series
  lStartValue := 0;

  lIdx := 0;
  lColorBand := GetColorBandTool(lStartValue);
  while assigned(lColorBand) do
  begin
    lAnnTool := GetAnnotationTool(GetAnnCaption(lIdx));

    if not assigned(lAnnTool) then
    begin
      sleep(1);
      lAnnTool := GetAnnotationTool(GetAnnCaption(lIdx));
      Assert(assigned(lAnnTool), Format('Band Tools: %d'#13'Annotation Tools: %d'#13'Captions: %s', [FBands.Count, FAnnotations.COunt, GetAnnoCaptions]));
    end;

    if assigned(lAnnTool) then
    begin
      UpdateAnnPos(lAnnTool, lColorBand);
      lStartValue := lColorBand.EndValue;
      lColorBand := GetColorBandTool(lStartValue);
      if assigned(FActiveColorBand) and (FActiveColorBand = lColorBand) then
        lColorBand := nil;
      inc(lIdx);
    end
    else
      lColorBand := nil;
  end;
end;

procedure TForm1.UpdateAnnPos(const aAnnTool: TAnnotationTool; const aColorBand: TColorBandTool);
var
  lSrs: TChartSeries;
  lLeft: Integer;
  lRight: Integer;
  lMiddle: Integer;
  lR: TRect;
  lX, lY: integer;
  lBtm: integer;
  lTop: Integer;
begin
  Assert(assigned(aAnnTool.ParentChart));
  Assert(aAnnTool.ParentChart.SeriesCount > 0);

  lSrs := aAnnTool.ParentChart.Series[0];
  aAnnTool.Visible := lSrs.ValuesList.Count > 0;
  if not aAnnTool.Visible  then
    exit;

  lLeft := lSrs.CalcXPosValue(aColorBand.StartValue);
  lRight := lSrs.CalcXPosValue(aColorBand.EndValue);

  lMiddle := lLeft + ((lRight - lLeft) div 2);


  lR := TAnnotationAccess(aAnnTool).GetTextBounds(lX, lY);

  aAnnTool.Shape.Left := lMiddle - (Max(0, lR.Right- lR.Left + 1) div 2);


  lTop := lSrs.CalcYPosValue(lSrs.ValuesList[1].MaxValue);
  lBtm := lSrs.CalcYPosValue(0);
  lMiddle := lBtm + ((lTop - lBtm) div 2);
  aAnnTool.Shape.Top := lMiddle - (Max(0, lR.Bottom - lR.Top + 1) div 2);

end;

function TForm1.CalcOffset(const index: Integer; aXPos: Integer): Double;
begin
  if Series1.Count > 1 then
    result:=(Series1.XValue[1]-Series1.XValue[0]) / 2
  else
    result:=0.5;

  if aXPos > Series1.CalcXPos(index) + Series1.BarWidth / 2 then
  begin
    if index < Series1.Count-1 then
      result:=(Series1.XValue[index+1]-Series1.XValue[index]) / 2
  end
  else if index > -1 then
    result:= - (Series1.XValue[index]-Series1.XValue[index-1]) / 2;
end;

procedure TForm1.Chart1BeforeDrawSeries(Sender: TObject);
begin
  Assert(Sender is TChart);
  UpdateAnnotations(TChart(Sender));
end;

end.
